import { always, isNil, useWith, prop, identity, path, append, curry, cond, curryN, concat, prepend, has, find, keys, test, isEmpty, T, __, tail, head, fromPairs, toPairs, pipe, replace } from 'ramda'
import isPath from './isPath'
import listPath from './listPath'

const getProp = useWith(path, [
  pipe(head, append(__, ['properties'])),
  identity
])

const matchPattern = (path, schema) => find(
  pipe(RegExp, test(__, head(path))),
  pipe(prop('patternProperties'), keys)(schema)
)

const appendRef = (ref, path) => concat(
  pipe(replace('#/definitions', ''), listPath)(ref),
  path
)

// newSchema :: object -> any
// o = {
//   xs: the current path
//   a: the current schema
//   b: the initial schema
// }
const newSchema = cond([

  // Check if the current path is present in the "properties" property
  [
    o => getProp(o.xs, o.a),
    o => ({
      xs: tail(o.xs),
      a: getProp(o.xs, o.a),
      b: o.b
    })
  ],

  // Check if the current path is present in the "patternPropperties" property
  [
    o => matchPattern(o.xs, o.a),
    o => ({
      xs: tail(o.xs),
      a: path(['patternProperties', matchPattern(o.xs, o.a)], o.a),
      b: o.b
    })
  ],

  // Check for reference to the definitions
  [
    o => has('$ref', o.a),
    o => ({
      xs: appendRef(prop('$ref', o.a), o.xs),
      a: { properties: o.b.definitions },
      b: o.b
    })
  ],

  // If none of the above has tested then something is not supported
  [T, always(undefined)]
])

const getSchema = pipe(newSchema, o => {
  return isNil(o) ? undefined :
    isEmpty(o.xs) ? o.a : getSchema(o)
})

/**
 * Given a JSON schema and a JSON pointer it gets
 * the value found at that pointer.
 *
 * Works for both properties and patternProperties
 *
 * https://tools.ietf.org/html/draft-zyp-json-schema-04
 * https://tools.ietf.org/html/rfc6901
 *
 * @function
 * @implements curry
 * @param {string} path - A JSON Path string
 * @param {object} schema - A JSON Schema object
 * @returns {any} - The value found at that path or undefined if nothing was found
 * @example
 *
 * import getSchemaPath from 'json-patch-utils'
 *
 * let schema = require('./schema.json')
 * schema.properties.foo.properties.bar.properties.baz // { type: "integer" }
 *
 * schemaPath('/foo/bar/baz', schema) // { type: "integer" }
 * schemaPath('/foo/BAM', schema) // undefined
 *
 */
const schemaPath = (xs, a) => getSchema({ xs: listPath(xs), a: a, b: a })

export default schemaPath
